/*
 * MikroSDK - MikroE Software Development Kit
 * Copyright (c) 2019, MikroElektronika - www.mikroe.com
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*!
 * \file
 *
 */

#include "magneto.h"

// ------------------------------------------------------------- PRIVATE MACROS 

#define MAGNETO_DUMMY                       0
                                            
#define MAGNETO_PARITY_CALIB_FACTOR         2
#define MAGNETO_START_VALUE                 0
#define MAGNETO_ANGLE_START_VALUE           0.0
#define MAGNETO_ANGLE_MULTIP_FACTOR         0.0219
#define MAGNETO_ANGLE_DIV_FACTOR            10.0
#define MAGNETO_BIT_0                       0x01
#define MAGNETO_STATE_LOW                   0

// -------------------------------------------------------------- PRIVATE TYPES


// ------------------------------------------------------------------ CONSTANTS


// ------------------------------------------------------------------ VARIABLES


// ---------------------------------------------- PRIVATE FUNCTION DECLARATIONS 

static uint8_t calc_parity ( magneto_t *ctx, uint16_t parity );

static void dev_comm_delay ( void );

// ------------------------------------------------ PUBLIC FUNCTION DEFINITIONS

void magneto_cfg_setup ( magneto_cfg_t *cfg )
{
    // Communication gpio pins 

    cfg->sck        = HAL_PIN_NC;
    cfg->miso       = HAL_PIN_NC;
    cfg->mosi       = HAL_PIN_NC;
    cfg->cs         = HAL_PIN_NC;

    // Additional gpio pins

    cfg->out        = HAL_PIN_NC;

    cfg->spi_speed  = SPI_MEDIUM; 
    cfg->spi_mode   = SPI_MODE_0;
}

MAGNETO_RETVAL magneto_init ( magneto_t *ctx, magneto_cfg_t *cfg )
{
    spi_master_config_t spi_cfg;

    spi_master_configure_default( &spi_cfg );
    spi_cfg.mode         = cfg->spi_mode;
    spi_cfg.speed        = cfg->spi_speed;
    spi_cfg.pins.sck     = cfg->sck;
    spi_cfg.pins.miso    = cfg->miso;
    spi_cfg.pins.mosi    = cfg->mosi;
    spi_cfg.chip_select  = cfg->cs;

    if (  spi_master_open( &ctx->spi, &spi_cfg ) != SPI_SUCCESS )
    {
        return MAGNETO_INIT_ERROR;
    }

    spi_master_set_dummy_data( &ctx->spi, MAGNETO_DUMMY );

    // Output pins 
    
    digital_out_init( &ctx->out, cfg->out );

    return MAGNETO_OK;

}

void magneto_generic_transfer ( magneto_t *ctx, spi_master_transfer_data_t *block )
{
    spi_master_start( &ctx->spi );
    dev_comm_delay( );
    spi_master_transfer( &ctx->spi, block );
    spi_master_stop( &ctx->spi );    
}

uint16_t magneto_read_data ( magneto_t *ctx, uint16_t address_command )
{
    spi_master_transfer_data_t block;
    
    uint8_t tx_buf[ 3 ];
    uint8_t rx_buf[ 4 ];
    uint16_t result;
    
    address_command  &= MAGNETO_REG_ANGLE;
    address_command  |= MAGNETO_CMD_READ_COMMAND_BIT;
    address_command  |= ( calc_parity( ctx, address_command ) << MAGNETO_CMD_PARITY_VALUE);
    
    tx_buf[ 0 ] = ( uint8_t ) ( address_command );
    tx_buf[ 0 ] &= ( uint8_t ) MAGNETO_CMD_HIGH_BYTE;
    tx_buf[ 0 ] >>= 8;
    tx_buf[ 1 ] = ( uint8_t ) ( address_command );
    tx_buf[ 1 ] &= ( uint8_t ) MAGNETO_CMD_LOW_BYTE;
    
    block.tx_buffer   = tx_buf;
    block.tx_length   = 2;
    block.rx_buffer   = rx_buf;
    block.rx_length   = 2;
    
    magneto_generic_transfer( ctx, &block );
    
    result = rx_buf[ 1 ];
    result <<= 8;
    result |= rx_buf[ 0 ];
    
    result &= MAGNETO_CMD_MAGNITUDE_VALUE;

    return result;
}

void magneto_write_data ( magneto_t *ctx, uint16_t address_command, uint16_t write_data )
{
    spi_master_transfer_data_t block;
    
    uint8_t tx_buf_cmd[ 2 ];
    uint8_t tx_buf_data[ 2 ];
    uint8_t rx_buf;

    address_command   &= MAGNETO_REG_ANGLE;
    address_command  |= ( calc_parity( ctx, address_command ) << MAGNETO_CMD_PARITY_VALUE);
    
    tx_buf_cmd[ 0 ]    = ( uint8_t ) ( address_command );
    tx_buf_cmd[ 0 ]   &= ( uint8_t ) MAGNETO_CMD_HIGH_BYTE;
    tx_buf_cmd[ 0 ]  >>= 8;
    tx_buf_cmd[ 1 ]    = ( uint8_t ) ( address_command );
    tx_buf_cmd[ 1 ]   &= ( uint8_t ) MAGNETO_CMD_LOW_BYTE;
    
    tx_buf_data[ 0 ]   = ( uint8_t ) ( address_command );
    tx_buf_data[ 0 ]  &= ( uint8_t ) MAGNETO_CMD_HIGH_BYTE;
    tx_buf_data[ 0 ] >>= 8;
    tx_buf_data[ 1 ]   = ( uint8_t ) ( write_data );
    tx_buf_data[ 1 ]  &= ( uint8_t ) MAGNETO_CMD_LOW_BYTE;
    
    block.tx_buffer = tx_buf_cmd;
    block.tx_length = 2;
    block.rx_buffer = &rx_buf;
    block.rx_length = 0;
    
    magneto_generic_transfer( ctx, &block );
    
    block.tx_buffer = tx_buf_data;
    block.tx_length = 2;
    block.rx_buffer = &rx_buf;
    block.rx_length = 0;
    
    magneto_generic_transfer( ctx, &block );
}

uint16_t magneto_get_state ( magneto_t *ctx )
{
    return magneto_read_data( ctx, MAGNETO_REG_DIAG_AGC );
}

uint8_t magneto_check_state ( magneto_t *ctx )
{
    if ( magneto_get_state( ctx ) != MAGNETO_STATE_LOW )
    {
        return MAGNETO_FLAG_ERROR;
    }
    else
    {
        return MAGNETO_FLAG_OK;
    }
}

float magneto_calculate_angle ( magneto_t *ctx )
{
    uint16_t dat_buf;
    uint8_t n_cnt;
    float angle;
    
    angle = MAGNETO_ANGLE_START_VALUE;
    
    magneto_read_data( ctx, MAGNETO_REG_CLEAR_ERROR_FLAG );
    
    magneto_write_data( ctx, MAGNETO_REG_OTP_ZERO_POS_HIGH, MAGNETO_CMD_RESET_VALUE );
    magneto_write_data( ctx, MAGNETO_REG_OTP_ZERO_POS_LOW, MAGNETO_CMD_RESET_VALUE );
    
    dat_buf   = magneto_read_data( ctx, MAGNETO_REG_MAGNITUDE );
    
    magneto_write_data( ctx, MAGNETO_REG_OTP_ZERO_POS_HIGH, ( ( dat_buf >> 6 ) & MAGNETO_CMD_SET_VALUE ) );
    magneto_write_data( ctx, MAGNETO_REG_OTP_ZERO_POS_LOW, ( dat_buf &  MAGNETO_REG_ANGLE ) );
    
    dat_buf   = magneto_get_state( ctx );

    for ( n_cnt = 0; n_cnt < 10; n_cnt++ )
    {
        dat_buf = magneto_read_data( ctx, MAGNETO_REG_MAGNITUDE );
        angle  += ( float ) ( dat_buf  * MAGNETO_ANGLE_MULTIP_FACTOR );
    }

    angle /= MAGNETO_ANGLE_DIV_FACTOR;
    
    return angle;
}

// ----------------------------------------------- PRIVATE FUNCTION DEFINITIONS

static uint8_t calc_parity ( magneto_t *ctx, uint16_t parity )
{
    uint8_t n_cnt;
    uint16_t result;
    
    result = MAGNETO_START_VALUE;

    for ( n_cnt = 0; n_cnt < 15; n_cnt++ )
    {
        result += ( parity >> n_cnt ) & MAGNETO_BIT_0;
        result &= MAGNETO_BIT_0;
    }
    
    result %= MAGNETO_PARITY_CALIB_FACTOR;
    
    return result;
}

static void dev_comm_delay ( void )
{
    Delay_1us();
}

// ------------------------------------------------------------------------- END

